@@ -1,15 +1,22 @@
+"""Contextual variables for assignations."""
+
 import contextvars
-from rekuest_next.actors.helper import AssignmentHelper
 from rekuest_next.actors.errors import (
     NotWithinAnAssignationError,
 )
+from typing import TYPE_CHECKING
+
+if TYPE_CHECKING:
+    from rekuest_next.actors.helper import AssignmentHelper
 
 
-current_assignment = contextvars.ContextVar("current_assignment")
-current_assignation_helper = contextvars.ContextVar("assignment_helper")
+current_assignation_helper: contextvars.ContextVar["AssignmentHelper"] = (
+    contextvars.ContextVar("assignment_helper")
+)
 
 
-def get_current_assignation_helper() -> AssignmentHelper:
+def get_current_assignation_helper() -> "AssignmentHelper":
+    """Get the current assignation helper."""
     try:
         return current_assignation_helper.get()
     except LookupError as e:
@@ -18,11 +25,11 @@ def get_current_assignation_helper() -> AssignmentHelper:
         ) from e
 
 
-def is_inside_assignation():
+def is_inside_assignation() -> bool:
     """Checks if the current context is inside an assignation (e.g. was called from
     the rekuest_server)"""
     try:
-        current_assignment.get()
+        current_assignation_helper.get()
         return True
     except LookupError:
         return False
